/*
 * README
 *
 * This is an example to show you how the APIs of SEkey can be used to build the KMS
 * and to manage the keys, the groups and the users. Remember that SEkey is not a "plug-and-play"
 * KMS, it is instead a framework with a collection of APIs that can be used to implement a KMS
 * according to the needs of the implementer. Pay attention to the comments in the code to
 * understand some key aspects of the SEcube platform.
*/

#include <fstream>
#include "../sources/L1/L1.h"
#include "../sekey/SEkey.h"

using namespace std;
void admin_mode(); // simulate admin
void user_mode(); // simulate user
extern userinfo currentuser; // useful for user sekey test

void print_db_content(); /* Example about how to print the content of the SEkey database of the current user (or the administrator). */
int init_user_secube_wrapper(std::string& uid, std::string& userpin, std::string& adminpin); /* Example about how to initialize the SEcube of a user. */

int main()
{
	string userinput;
    cout << "press 1 for admin mode" << endl;
    cout << "press 2 for user mode" << endl;
    cout << "press q to quit" << endl;
    getline(cin, userinput);
    char ch = userinput.at(0);
    /* this example allows to run a simple SEkey implementation as administrator or user. notice that the specific SEcube
     * of the administrator or of a user must be used...otherwise the system will not work. */
    switch(ch){
    case '1':
    	admin_mode();
    	break;
    case '2':
    	user_mode();
    	break;
    case 'q':
    	return 0;
    default:
    	cout << "Invalid choice." << endl;
    }
    cout << "Press q to quit the program." << endl;
    while(1){
		getline(cin, userinput);
    	if(userinput.compare("q") == 0){
    		break;
    	}
    }
    return 0;
}

int init_user_secube_wrapper(string& uid, string& userpin, string& adminpin){
	vector<string> pins; // vector of pins to try for login as admin on user secube
	pins.push_back(string("user"));
	pins.push_back(string("admin"));
	pins.push_back(string(32, 0)); // this is the default pin when the SEcube has to be initialized
	/* warning: the pin must be provided to the L1Login API as an uint8_t[] so the string must be converted to
	 * a raw array of bytes. this means that if the pin is all zeroes, the corresponding string can't be
	 * initialized as "000...0" because those zeroes would then be converted from ASCII to normal binary
	 * end each byte would be = 48 (decimal). */
	cout << "Please connect the SEcube of the user to the host machine. Press 'n' when ready." << endl;
	/* WARNING FOR LINUX USERS: when you connect the SEcube of the user to the host computer, remember to mount the SEcube otherwise this API might fail. */
	string userinput;
	getline(cin, userinput);
	if(userinput.compare("n") == 0){
		int x = sekey_init_user_SEcube(uid, userpin, adminpin, pins);
		while(1){
			cout << "Please disconnect the SEcube of the user from the host machine. Press 'n' when done." << endl;
			getline(cin, userinput);
			if(userinput.compare("n") == 0){
				L0 l0;
				if(l0.GetNumberDevices() == 1){
					return x;
				}
			}
		}
	} else {
		return -1;
	}
}

void print_db_content(){
	ofstream sekey_log;
	string fname;
	fname = currentuser.userid + '_' + "dump.txt";
	sekey_log.open (fname.c_str());
	// print list of users
	vector<se_user> users;
	if(sekey_user_get_info_all(&users) != 0){
		return;
	}
	for(se_user& user : users){
		user.print_user_details(sekey_log);
	}
	// print list of groups
	vector<se_group> groups;
	if(sekey_group_get_info_all(&groups) != 0){
		return;
	}
	for(se_group& group : groups){
		group.print_group_details(sekey_log);
	}
	// print list of keys
	vector<se_key> keys;
	if(sekey_key_get_info_all(&keys) != 0){
		return;
	}
	for(se_key& key : keys){
		key.print_key_details(sekey_log);
	}
	sekey_log.close();
	return;
}

/* this is to simulate the actions of the KMS administrator */
void admin_mode(){

	/* this is a list of pins that may be used to perform the login (with admin or user privilege) on any SEcube.
	 * notice that in a real-world scenario the software should ask for the pin to the user of the PC (by means
	 * of a GUI or a CLI); here we use a hard-coded list for simplicity.  */
	vector<string> pins;
	pins.push_back(string("user"));
	pins.push_back(string("admin"));
	pins.push_back(string(32, 0)); // this is the default pin when the SEcube has been erased and re-programmed (all 0s)
	/* warning: the pin must be provided to the L1Login API as an uint8_t[] so the string must be converted to
	 * a raw array of bytes. this means that if the pin is all zeroes, the corresponding string can't be
	 * initialized as "000...0" because those zeroes would then be converted from ASCII to normal binary
	 * end each byte would be = 48 (decimal). */

	/* here we just decide the PINs that will be set for every SEcube that will be initialized (after erase and re-program).
	 * notice that we use dumb PINs for sake of simplicity, the real application should not use hard-coded PINs. it should
	 * generate the PINs automatically as a random 32 byte string or it should ask to the administrator to input the PINs
	 * on the keyboard. notice also that every SEcube must be initialized with an "admin PIN" and a "user PIN", these values
	 * must be different for every SEcube. the administrator of the KMS must keep track of the PINs of all SEcube(s), on the
	 * contrary the users are provided exclusively with the "user PIN" of their SEcube. */
	string userpin = "user";
	string adminpin = "admin";

    uint8_t numDevices = 0; // this is the number of SEcube(s) connected to the PC
	uint8_t pin_admin[32]; // this is the PIN to be used with L1Login API to login to the SEcube of the administrator (with admin privilege)
	memset(pin_admin, 0, 32); // we clean the PIN, then we set it to the dumb "admin" value
	memcpy(pin_admin, adminpin.c_str(), adminpin.length());

	/* do not mind to these variables...they are here simply to run the example */
	string gid, gname, uid, uname, kid, kname, h1, h2;
	int iteration = 0;
	group_policy gpolicy;
	sekey_error rc;

	/* L0 objects are required to communicate at the low-level (i.e. data over USB cable) with the SEcube.
	 * Notice that a smart pointer is used to incapsulate the L0 object, this is useful to avoid memory
	 * leakage that may happen if using the native object (you can use a unique pointer or a shared pointer
	 * depending on the context). */
	unique_ptr<L0> l0 = make_unique<L0>();
	numDevices = l0->GetNumberDevices(); // check how many SEcube(s) are connected
	if(numDevices==0){
		cout << "SEcube not connected. Exit." << endl;
		return;
	}

	/* L1 objects are used to perform high-level operations with the SEcube. You need a L1 object for every SEcube
	 * that is connected. Use the L1 object incapsulated by a unique ptr (or shared ptr) because the smart pointer
	 * allows to call automatically the destructor of the L1 object, whenever the L1 object must be "re-created" in
	 * situations like when the SEcube is removed and reconnected to the PC. */
	unique_ptr<L1> l1 = make_unique<L1>();

	/* notice that in this example we try to login to the SEcube of the admin with the "admin" PIN. if this does not work
	 * then it means that the SEcube of the administrator still needs to be initialized. */
	try{
		l1->L1Login(pin_admin, SE3_ACCESS_ADMIN, true); // if the login fails (i.e. wrong pin or secube not initialized), L1Login throws an exception
		cout << "SEcube logged in." << endl;
	} catch(...){
		try{
			while(1){
				/* this is an example about how to initialize the SEcube of the administrator. notice that the interaction with the admin, using the host PC,
				 * should be done according to the software you want to develop (i.e. GUI or CLI). here we simply print questions on the terminal; you should
				 * follow the same pattern in your software (i.e. add a button to the GUI to initialize the SEcube of the administrator). */
				cout << "The SEcube doesn't seem to be initialized. Do you want to initialize it as administrator SEcube for SEkey? (y/n)" << endl;
				string userinput;
				getline(cin, userinput);
				if(userinput.compare("y") == 0){
					/* here we call the API to initialize the SEcube of the administrator. notice that we pass the L1 object and the PINs to be set. */
					if(sekey_admin_init(*l1, pins, userpin, adminpin) != SEKEY_OK){
						/* pay attention to how failure is managed. simply return asking to erase again the SEcube (internal memory may be dirty after failure), flashing again the firmware and retrying to initialize the SEcube. */
						cout << "An error occured while initializing the SEcube. Please erase the SEcube and flash again the firmware, then retry." << endl;
						return;
					} else {
						cout << "Admin initialization completed successfully" << endl;
						/* notice how we reset the smart pointer of L0 and L1. this must be done because, after the initialization, the SEcube must be disconnected
						 * from the PC and reconnected. in order to avoid memory leakage L0 and L1 destructors must be called (this is automatic with smart ptrs). */
						l1.reset(); // L1 destructor is called
						l0.reset(); // L0 destructor is called
						while(1){
							string usrinput;
							cout << "Please disconnect and reconnect the administrator SEcube...press 'n' then enter when done.";
							getline(cin, usrinput);
							if(usrinput.compare("n") == 0){
								l0 = make_unique<L0>(); // notice how we create again the L0 object (incapsulated by a unique ptr)
								if(l0->GetNumberDevices() == 1){ // again, check the number of devices
									break;
								} else {
									l0.reset(); // if the number of devices is not exactly 1 (meaning no secube connected or too many secube connected) reset the pointer in order to call the L0 destructor
								}
							}
						}
						l1 = make_unique<L1>(); // L1 must be recreated because the SEcube was disconnected and reconnected, same pattern of L0
						l1->L1Login(pin_admin, SE3_ACCESS_ADMIN, true); // now we can login with the admin PIN that we set
						break;
					}
				}
				if(userinput.compare("n") == 0){
					cout << "Quit because the SEcube is not initialized." << endl;
					return;
				}
			}
		} catch (...){ // same as before...if something goes wrong erase the SEcube (because the flash memory may be dirty) and try again.
			cout << "An error occured while initializing the SEcube. Please erase the SEcube and flash again the firmware, then retry." << endl;
			return;
		}
	}

	/* Start SEkey, providing the required L0 and L1 objects (they are used as global variables by the APIs) */
	if(sekey_start(*l0, l1.get()) != 0){
		cout << "Error starting SEkey." << endl;
		return;
	}
	cout << "SEkey started." << endl;


	/* Now we will perform several operations on SEkey in order to show how the KMS works.
	 * Each operation is meant to test a specific API, the order of the operations is important
	 * to check that the system works properly. Some operations are intentionally wrong (i.e. wrong
	 * parameters or changes not allowed like activating a compromised key), after each API call the
	 * database content is printed to an output .txt file so that you can check it out.
	 * The return value of each API is checked against the expected value, if the returned value does not
	 * match, a log message is printed on the console. */

	/* ADD GROUPS 1,2,3,4,5 */
	iteration = 0;
	vector<pair<string,string>> groups;
	groups.push_back(pair<string,string>("G1", "Group_1")); // notice how we send hard-coded parameters to the APIs...this is just to simplify things
	groups.push_back(pair<string,string>("G2", "Group_2"));
	groups.push_back(pair<string,string>("G3", "Group_3"));
	groups.push_back(pair<string,string>("G4", "Group_4"));
	groups.push_back(pair<string,string>("G5", "Group_5"));
	for(pair<string,string> p : groups){
		gid = p.first;
		gname = p.second;
		gpolicy = group_policy(1000, L1Algorithms::Algorithms::AES_HMACSHA256, 31536000); // check out the meaning of policy
		rc = (sekey_error)sekey_add_group(gid, gname, gpolicy); // API
		if(rc != SEKEY_OK){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* ADD USERS 1,2,3,4,5 */
	iteration = 0;
	vector<pair<string,string>> users;
	users.push_back(pair<string,string>("U1", "User_1"));
	users.push_back(pair<string,string>("U2", "User_2"));
	users.push_back(pair<string,string>("U3", "User_3"));
	users.push_back(pair<string,string>("U4", "User_4"));
	users.push_back(pair<string,string>("U5", "User_5"));
	for(pair<string,string> p : users){
		uid = p.first;
		uname = p.second;
		rc = (sekey_error)sekey_add_user(uid, uname);
		if(rc != SEKEY_OK){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* WARNING: comment the next 4 API calls if you do not have the SEcube(s) to be initialized. */
	/* INITIALIZE USER 1 */
	uid = "U1";
	rc = (sekey_error)init_user_secube_wrapper(uid, userpin, adminpin);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	} else {
		cout << uid << " initialized." << endl;
	}
	/* INITIALIZE USER 2 */
	uid = "U2";
	rc = (sekey_error)init_user_secube_wrapper(uid, userpin, adminpin);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	} else {
		cout << uid << " initialized." << endl;
	}
	/* INITIALIZE USER 3 */
	uid = "U3";
	rc = (sekey_error)init_user_secube_wrapper(uid, userpin, adminpin);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	} else {
		cout << uid << " initialized." << endl;
	}
	/* INITIALIZE USER 4 */
	uid = "U4";
	rc = (sekey_error)init_user_secube_wrapper(uid, userpin, adminpin);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	} else {
		cout << uid << " initialized." << endl;
	}

	/* ADD USERS TO GROUPS */
	iteration = 0;
	vector<pair<string,string>> usergroups;
	usergroups.push_back(pair<string,string>("U1", "G1"));
	usergroups.push_back(pair<string,string>("U2", "G1"));
	usergroups.push_back(pair<string,string>("U3", "G2"));
	usergroups.push_back(pair<string,string>("U4", "G2"));
	usergroups.push_back(pair<string,string>("U5", "G3"));
	usergroups.push_back(pair<string,string>("U1", "G3"));
	usergroups.push_back(pair<string,string>("U4", "G3"));
	for(pair<string,string> p : usergroups){
		uid = p.first;
		gid = p.second;
		rc = (sekey_error)sekey_add_user_group(uid, gid);
		if(rc != SEKEY_OK){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* CREATE KEYS */
	iteration = 0;
	vector<pair<int,string>> keys;
	keys.push_back(pair<int,string>(101, "G1"));
	keys.push_back(pair<int,string>(106, "G1"));
	keys.push_back(pair<int,string>(107, "G1"));
	keys.push_back(pair<int,string>(108, "G1"));
	keys.push_back(pair<int,string>(102, "G2"));
	keys.push_back(pair<int,string>(109, "G2"));
	keys.push_back(pair<int,string>(110, "G2"));
	keys.push_back(pair<int,string>(103, "G3"));
	keys.push_back(pair<int,string>(111, "G3"));
	keys.push_back(pair<int,string>(112, "G3"));
	keys.push_back(pair<int,string>(113, "G3"));
	keys.push_back(pair<int,string>(104, "G4"));
	keys.push_back(pair<int,string>(114, "G4"));
	keys.push_back(pair<int,string>(115, "G4"));
	keys.push_back(pair<int,string>(105, "G5"));
	keys.push_back(pair<int,string>(116, "G5"));
	for(pair<int,string> p : keys){
		kid = "K" + to_string(p.first);
		kname = "Key_" + to_string(p.first);
		gid = p.second;
		rc = (sekey_error)sekey_add_key(kid, kname, gid, 0, se_key_type::symmetric_data_encryption);
		if(rc != SEKEY_OK){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* ACTIVATE KEYS */
	iteration = 0;
	vector<int> toactivate = {101, 109, 111, 115, 105};
	for(int k : toactivate){
		kid = "K" + to_string(k);
		rc = (sekey_error)sekey_activate_key(kid);
		if(rc != SEKEY_OK){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* DESTROY KEYS */
	iteration = 0;
	vector<int> todestroy = {116, 106, 113};
	for(int k : todestroy){
		kid = "K" + to_string(k);
		rc = (sekey_error)sekey_key_change_status(kid, se_key_status::destroyed);
		if(rc != SEKEY_OK){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* COMPROMISE KEYS */
	iteration = 0;
	vector<int> tocomp = {101, 105};
	for(int k : tocomp){
		kid = "K" + to_string(k);
		rc = (sekey_error)sekey_key_change_status(kid, se_key_status::compromised);
		if(rc != SEKEY_OK){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* DEACTIVATE KEYS */
	iteration = 0;
	vector<int> todeactivate = {109, 115};
	for(int k : todeactivate){
		kid = "K" + to_string(k);
		rc = (sekey_error)sekey_key_change_status(kid, se_key_status::deactivated);
		if(rc != SEKEY_OK){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* SUSPEND KEY 111 */
	kid = "K111";
	rc = (sekey_error)sekey_key_change_status(kid, se_key_status::suspended);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* ACTIVATE KEYS */
	iteration = 0;
	toactivate.clear();
	toactivate = {107, 108, 110, 103, 114};
	for(int k : toactivate){
		kid = "K" + to_string(k);
		rc = (sekey_error)sekey_activate_key(kid);
		if(rc != SEKEY_OK){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* SUSPEND KEY 107 */
	kid = "K107";
	rc = (sekey_error)sekey_key_change_status(kid, se_key_status::suspended);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* ACTIVATE KEY 111 */
	kid = "K111";
	rc = (sekey_error)sekey_activate_key(kid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
	}

	/* ACTIVATE KEY 115 (already deactivated) */
	kid = "K115";
	rc = (sekey_error)sekey_activate_key(kid);
	if(rc != SEKEY_ERR){
		cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
	}

	/* ACTIVATE KEY 105 (already compromised) */
	kid = "K105";
	rc = (sekey_error)sekey_activate_key(kid);
	if(rc != SEKEY_ERR){
		cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
	}

	/* ACTIVATE KEY 116 (already destroyed) */
	kid = "K116";
	rc = (sekey_error)sekey_activate_key(kid);
	if(rc != SEKEY_ERR){
		cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
	}

	/* SET TO KEYS PREACTIVE */
	iteration = 0;
	vector<int> topreactive = {105, 107, 108};
	for(int k : topreactive){
		kid = "K" + to_string(k);
		rc = (sekey_error)sekey_key_change_status(kid, se_key_status::preactive);
		if(rc != SEKEY_ERR_PARAMS){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* CHANGE KEYS NAME */
	iteration = 0;
	vector<int> torename = {101, 102, 103, 104, 105, 106, 107, 108, 109, 110};
	for(int k : torename){
		kid = "K" + to_string(k);
		kname = "KEY" + to_string(k);
		rc = (sekey_error)sekey_key_change_name(kid, kname);
		if(rc != SEKEY_OK){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* CHANGE GROUPS NAME */
	iteration = 0;
	torename.clear();
	torename = {1, 2, 3, 4};
	for(int k : torename){
		gid = "G" + to_string(k);
		gname = "GROUP" + to_string(k);
		rc = (sekey_error)sekey_group_change_name(gid, gname);
		if(rc != SEKEY_OK){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* CHANGE GROUPS MAX KEYS */
	iteration = 0;
	torename.clear();
	torename = {1, 2, 3, 4};
	for(int k : torename){
		gid = "G" + to_string(k);
		rc = (sekey_error)sekey_group_change_max_keys(gid, 9999);
		if(rc != SEKEY_OK){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* CHANGE GROUPS LIVENESS */
	iteration = 0;
	torename.clear();
	torename = {1, 2, 3, 4};
	for(int k : torename){
		gid = "G" + to_string(k);
		rc = (sekey_error)sekey_group_change_default_cryptoperiod(gid, 2629800);
		if(rc != SEKEY_OK){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* CHANGE USER NAMES */
	iteration = 0;
	torename.clear();
	torename = {1, 2, 3};
	for(int k : torename){
		uid = "U" + to_string(k);
		uname = "USER" + to_string(k);
		rc = (sekey_error)sekey_user_change_name(uid, uname);
		if(rc != SEKEY_OK){
			cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
		}
		iteration++;
	}

	/* ADD USER 5 GROUP 2 */
	uid = "U5";
	gid = "G2";
	rc = (sekey_error)sekey_add_user_group(uid, gid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* REMOVE USER 1 GROUP 3 */
	uid = "U1";
	gid = "G3";
	rc = (sekey_error)sekey_delete_user_group(uid, gid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* REMOVE USER 5 GROUP 2 */
	uid = "U5";
	gid = "G2";
	rc = (sekey_error)sekey_delete_user_group(uid, gid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* REMOVE USER 4 GROUP 4 */
	uid = "U4";
	gid = "G4";
	rc = (sekey_error)sekey_delete_user_group(uid, gid);
	if(rc != SEKEY_UNCHANGED){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* REMOVE USER 2 GROUP 2 */
	uid = "U2";
	gid = "G2";
	rc = (sekey_error)sekey_delete_user_group(uid, gid);
	if(rc != SEKEY_UNCHANGED){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* DELETE USER 1 */
	uid = "U1";
	rc = (sekey_error)sekey_delete_user(uid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* DELETE USER 1 (already deleted) */
	uid = "U1";
	rc = (sekey_error)sekey_delete_user(uid);
	if(rc != SEKEY_USER_NOT_FOUND){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* DELETE USER THAT DOES NOT EXIST */
	uid = "U123456";
	rc = (sekey_error)sekey_delete_user(uid);
	if(rc != SEKEY_USER_NOT_FOUND){
		cout << "Error at line " << __LINE__ << endl;
		cin.get();
	}

	/* DELETE GROUP 4 */
	gid = "G4";
	rc = (sekey_error)sekey_delete_group(gid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* DELETE GROUP 5 */
	gid = "G5";
	rc = (sekey_error)sekey_delete_group(gid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* RECREATE USER 1 */
	uid = "U1";
	uname = "USER_1";
	rc = (sekey_error)sekey_add_user(uid, uname);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* ADD USER 1 TO GROUP 1 */
	uid = "U1";
	gid = "G1";
	rc = (sekey_error)sekey_add_user_group(uid, gid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	// change name of user that does not exist
	uid = "U123456";
	uname = "user_123456";
	rc = (sekey_error)sekey_user_change_name(uid, uname);
	if(rc != SEKEY_USER_NOT_FOUND){
		cout << "Error at line " << __LINE__ << endl;
		cin.get();
	}

	/* ADD USER 1 TO GROUP 2 */
	uid = "U1";
	gid = "G2";
	rc = (sekey_error)sekey_add_user_group(uid, gid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* DELETE USER 2 */
	uid = "U2";
	rc = (sekey_error)sekey_delete_user(uid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* FIND KEY FOR USER 3 AND USER 4 */
	string source, dest, chosen;
	source = "U3";
	dest = "U4";
	rc = (sekey_error)sekey_find_key_v1(chosen, source, dest, se_key_type::symmetric_data_encryption);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	} else {
		cout << "Key for " + source + ", " + dest + ": " + chosen << endl;
	}

	/* FIND KEY FOR USER 1 AND USER 5 */
	source = "U1";
	dest = "U5";
	rc = (sekey_error)sekey_find_key_v1(chosen, source, dest, se_key_type::symmetric_data_encryption);
	if(rc == SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
		cout << "Key for " + source + ", " + dest + ": " + chosen << endl;
	} else {
		cout << "Key for " + source + ", " + dest + " not found. Returned value: " << rc << endl;
	}

	/* FIND KEY FOR USER 1 AND USER 3 */
	source = "U1";
	dest = "U3";
	rc = (sekey_error)sekey_find_key_v1(chosen, source, dest, se_key_type::symmetric_data_encryption);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	} else {
		cout << "Key for " + source + ", " + dest + ": " + chosen << endl;
	}

	/* ADD ANOTHER ACTIVE KEY TO GROUP 2 */
	kid = "K231";
	kname = "Key_231";
	gid = "G2";
	rc = (sekey_error)sekey_add_key(kid, kname, gid, 1440, se_key_type::symmetric_data_encryption);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
	}
	rc = (sekey_error)sekey_activate_key(kid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
	}

	/* FIND KEY FOR USER 4 AND USER 5 */
	source = "U4";
	dest = "U5";
	rc = (sekey_error)sekey_find_key_v1(chosen, source, dest, se_key_type::symmetric_data_encryption);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	} else {
		cout << "Key for " + source + ", " + dest + ": " + chosen << endl;
	}

	/* ADD USER 3 TO GROUP 3 */
	uid = "U3";
	gid = "G3";
	rc = (sekey_error)sekey_add_user_group(uid, gid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* FIND KEY FOR USER 3 AND USER 4 */
	source = "U3";
	dest = "U4";
	rc = (sekey_error)sekey_find_key_v1(chosen, source, dest, se_key_type::symmetric_data_encryption);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	} else {
		cout << "Key for " + source + ", " + dest + ": " + chosen << endl;
	}

	/* ADD USER 5 TO GROUP 2 */
	uid = "U5";
	gid = "G2";
	rc = (sekey_error)sekey_add_user_group(uid, gid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* FIND KEY FOR USER 3 AND USER 4 */
	source = "U3";
	dest = "U4";
	rc = (sekey_error)sekey_find_key_v1(chosen, source, dest, se_key_type::symmetric_data_encryption);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	} else {
		cout << "Key for " + source + ", " + dest + ": " + chosen << endl;
	}

	/* ADD GROUP 4 */
	gid = "G4";
	gname = "GROUP_4";
	gpolicy = group_policy(1000, L1Algorithms::Algorithms::AES_HMACSHA256, 31536000);
	rc = (sekey_error)sekey_add_group(gid, gname, gpolicy);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
	}

	/* ADD USER 1 TO GROUP 4 */
	uid = "U1";
	gid = "G4";
	rc = (sekey_error)sekey_add_user_group(uid, gid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* ADD USER 2 TO GROUP 4 */
	uid = "U2";
	gid = "G4";
	rc = (sekey_error)sekey_add_user_group(uid, gid);
	if(rc == SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* RECREATE USER 2 */
	uid = "U2";
	uname = "USER_2";
	rc = (sekey_error)sekey_add_user(uid, uname);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* ADD USER 2 TO GROUP 4 */
	uid = "U2";
	gid = "G4";
	rc = (sekey_error)sekey_add_user_group(uid, gid);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	/* ADD TWO KEYS TO GROUP 4 */
	kid = "K401";
	kname = "Key_401";
	gid = "G4";
	rc = (sekey_error)sekey_add_key(kid, kname, gid, 1440, se_key_type::symmetric_data_encryption);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
	}
	kid = "K402";
	kname = "Key_402";
	gid = "G4";
	rc = (sekey_error)sekey_add_key(kid, kname, gid, 1440, se_key_type::symmetric_data_encryption);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << " during iteration " << iteration << endl;
	}

	/* FIND KEY FOR USER 1 AND USER 2 */
	source = "U1";
	dest = "U2";
	rc = (sekey_error)sekey_find_key_v1(chosen, source, dest, se_key_type::symmetric_data_encryption);
	if(rc == SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
		cout << "Key for " + source + ", " + dest + ": " + chosen << endl;
	} else {
		cout << "Key for " + source + ", " + dest + " not found. Returned value: " << rc << endl;
	}

	/* TEST KEY GET INFO API */
	kid = "K231";
	se_key k;
	if(sekey_key_get_info(kid, &k) != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	} else {
		ofstream keyinfo("keygetinfo.txt");
		k.print_key_details(keyinfo);
		keyinfo.close();
	}

	/* TEST GROUP GET INFO API */
	gid = "G3";
	se_group g;
	if(sekey_group_get_info(gid, &g) != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	} else {
		ofstream groupinfo("groupgetinfo.txt");
		g.print_group_details(groupinfo);
		groupinfo.close();
	}

	/* FIND KEY FOR USER 1 AND GROUP 2 */
	source = "U1";
	dest = "G2";
	rc = (sekey_error)sekey_find_key_v2(chosen, source, dest, se_key_type::symmetric_data_encryption);
	if(rc == SEKEY_OK){
		cout << "Key for " + source + ", " + dest + ": " + chosen << endl;
	} else {
		cout << "Key for " + source + ", " + dest + " not found. Returned value: " << rc << endl;
	}

	/* FIND KEY FOR USER 1 AND GROUP 9 */
	source = "U1";
	dest = "G9";
	rc = (sekey_error)sekey_find_key_v2(chosen, source, dest, se_key_type::symmetric_data_encryption);
	if(rc == SEKEY_OK){
		cout << "Key for " + source + ", " + dest + ": " + chosen << endl;
	} else {
		cout << "Key for " + source + ", " + dest + " not found. Returned value: " << rc << endl;
	}

	/* FIND KEY FOR USER 1 AND GROUP 3 */
	source = "U1";
	dest = "G3";
	rc = (sekey_error)sekey_find_key_v2(chosen, source, dest, se_key_type::symmetric_data_encryption);
	if(rc == SEKEY_OK){
		cout << "Key for " + source + ", " + dest + ": " + chosen << endl;
	} else {
		cout << "Key for " + source + ", " + dest + " not found. Returned value: " << rc << endl;
	}

	/* FIND KEY FOR USER 10 AND GROUP 2 */
	source = "U10";
	dest = "G2";
	rc = (sekey_error)sekey_find_key_v2(chosen, source, dest, se_key_type::symmetric_data_encryption);
	if(rc == SEKEY_OK){
		cout << "Key for " + source + ", " + dest + ": " + chosen << endl;
	} else {
		cout << "Key for " + source + ", " + dest + " not found. Returned value: " << rc << endl;
	}

	/* FIND KEY FOR USER 3 AND USERS 4,5 */
	source = "U3";
	vector<string> dest_;
	dest_.push_back("U4");
	dest_.push_back("U5");
	rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
	if(rc == SEKEY_OK){
		cout << "Key for " + source + " multicast: " + chosen << endl;
	} else {
		cout << "Key for " + source + " multicast not found. Returned value: " << rc << endl;
	}

	/* FIND KEY FOR USER 3 AND USERS 4,1 */
	source = "U3";
	dest_.clear();
	dest_.push_back("U4");
	dest_.push_back("U1");
	rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
	if(rc == SEKEY_OK){
		cout << "Key for " + source + " multicast: " + chosen << endl;
	} else {
		cout << "Key for " + source + " multicast not found. Returned value: " << rc << endl;
	}

	/* FIND KEY FOR USER 1 AND USERS 2,5 */
	source = "U1";
	dest_.clear();
	dest_.push_back("U2");
	dest_.push_back("U5");
	rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
	if(rc == SEKEY_OK){
		cout << "Key for " + source + " multicast: " + chosen << endl;
	} else {
		cout << "Key for " + source + " multicast not found. Returned value: " << rc << endl;
	}

	/* FIND KEY FOR USER 1 AND USERS 2,5 */
	source = "U9";
	dest_.clear();
	dest_.push_back("U2");
	dest_.push_back("U5");
	rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
	if(rc == SEKEY_OK){
		cout << "Key for " + source + " multicast: " + chosen << endl;
	} else {
		cout << "Key for " + source + " multicast not found. Returned value: " << rc << endl;
	}

	/* FIND KEY FOR USER 1 AND USERS 2,5 */
	source = "U1";
	dest_.clear();
	dest_.push_back("U2");
	dest_.push_back("U8");
	rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
	if(rc == SEKEY_OK){
		cout << "Key for " + source + " multicast: " + chosen << endl;
	} else {
		cout << "Key for " + source + " multicast not found. Returned value: " << rc << endl;
	}

	/* FIND KEY FOR USER 1 AND USERS 2,5 */
	source = "U1";
	dest_.clear();
	dest_.push_back("U3");
	dest_.push_back("U4");
	dest_.push_back("U5");
	rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
	if(rc == SEKEY_OK){
		cout << "Key for " + source + " multicast: " + chosen << endl;
	} else {
		cout << "Key for " + source + " multicast not found. Returned value: " << rc << endl;
	}

	/* REQUEST RECOVERY FOR USER 4 */
	se_user su;
	uid = "U4";
	if(sekey_user_get_info(uid, &su) != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}
	string sn = su.get_sn();
	rc = (sekey_error)sekey_recovery_request(uid, sn);
	if(rc != SEKEY_OK){
		cout << "Error at line " << __LINE__ << endl;
	}

	// end the admin mode
	sekey_stop();
	cout << "SEkey stopped." << endl;
	l1->L1Logout();
	return;
}

/* here we simulate the behavior of a user */
void user_mode(){

	uint8_t numDevices = 0; // how many SEcube(s) are connected
	uint8_t pin_user[32]; // to login on the SEcube as user
	memset(pin_user, 0, 32);
	memcpy(pin_user, "user", 4);

	/* these variables are used only to run the example */
	string gid, gname, uid, uname, kid, kname;
	group_policy gpolicy;
	sekey_error rc;

	unique_ptr<L0> l0 = make_unique<L0>(); // use smart ptr to avoid memory leakage (check comment on admin function above)
	numDevices = l0->GetNumberDevices(); // count how many SEcube(s) are connected
	if(numDevices==0){
		cout << "No SEcube devices connected. Press a key to quit." << endl;
		return;
	}

	unique_ptr<L1> l1 = make_unique<L1>(); // use smart ptr to avoid memory leakage (check comment on admin function above)
	try{
		l1->L1Login(pin_user, SE3_ACCESS_USER, true);
	} catch(...){
		cout << "The SEcube doesn't seem to be initialized, therefore it can't be used with SEkey. Please give the SEcube to the"
				"SEkey administrator in order to initialize it, then launch this program again." << endl;
		return;
	}

	if(sekey_start(*l0, l1.get()) != 0){ // start SEkey
		cout << "Error starting SEkey." << endl;
		return;
	}

	/* explicit request to update the data of SEkey for the user (you don't really need to do this, it is already done automatically) */
	rc = (sekey_error)sekey_update_userdata();
	if(rc == SEKEY_OK){
		cout << "SEkey is up to date!" << endl;
		print_db_content();
	} else {
		cout << "Update failed! SEkey is blocked until the update will be completed successfully." << endl;
	}

	// don't mind to these lines
	string source = currentuser.userid;
	vector<se_user> users;
	string dest, h1, h2, chosen;

	/* TEST KEY GET INFO API: we test both APIs with a trick, first we retrieve all info of all keys with the sekey_key_get_info_all(), then we
	 * use only the IDs of the keys to retrieve the info of each specific key. This is not required in the real-world application, if you already
	 * know the ID of the key that you need, simply use sekey_key_get_info(). Notice that sekey_key_get_info_all() is used when you want to
	 * retrieve general info or when you want to perform some kind of aggregation on the data (i.e. counting how many keys there are, counting how
	 * many keys belong to a certain group, counting the number of active keys, etc.). */
	vector<se_key> keys;
	if(sekey_key_get_info_all(&keys) != 0){
		cout << "Error retrieving all keys." << endl;
	}
	for(se_key& key : keys){
		se_key k;
		if(sekey_key_get_info(key.get_id(), &k) != SEKEY_OK){
			cout << "Error at line " << __LINE__ << endl;
		} else {
			ofstream keyinfo("keygetinfo.txt", fstream::out | fstream::app);
			k.print_key_details(keyinfo);
			keyinfo.close();
		}
	}

	/* TEST GROUP GET INFO API. Notice that we use the same trick explained above for the keys. */
	vector<se_group> groups;
	if(sekey_group_get_info_all(&groups) != 0){
		cout << "Error retrieving all groups." << endl;
	}
	for(se_group& gr : groups){
		se_group g;
		if(sekey_group_get_info(gr.get_id(), &g) != SEKEY_OK){
			cout << "Error at line " << __LINE__ << endl;
		} else {
			ofstream groupinfo("groupgetinfo.txt", fstream::out | fstream::app);
			g.print_group_details(groupinfo);
			groupinfo.close();
		}
		dest = g.get_id();
		chosen.clear();
		/* here we test the API that retrieves the most secure key to be used in a communication between the user
		 * and the members of a group */
		rc = (sekey_error)sekey_find_key_v2(chosen, source, dest, se_key_type::symmetric_data_encryption);
		if(rc == SEKEY_OK){
			cout << "Key for " + source + ", " + dest + ": " + chosen << endl;
		} else {
			cout << "Key for " + source + ", " + dest + " not found. Returned value: " << rc << endl;
		}
	}

	/* TEST USER GET INFO API. Same trick used for keys APIs above. */
	vector<se_user> users_;
	if(sekey_user_get_info_all(&users_) != 0){
		cout << "Error retrieving all users." << endl;
	}
	for(se_user& us : users_){
		se_user u;
		if(sekey_user_get_info(us.get_id(), &u) != SEKEY_OK){
			cout << "Error at line " << __LINE__ << endl;
		} else {
			ofstream userinfo("usergetinfo.txt", fstream::out | fstream::app);
			u.print_user_details(userinfo);
			userinfo.close();
		}
	}

	/* FIND KEY FOR EACH COUPLE OF USERS. First we retrieve the list of users we know, then we search for the most secure
	 * key to be used with each specific user. */
	if(sekey_user_get_info_all(&users) != 0){
		cout << "Error retrieving all users." << endl;
	}
	for(se_user& user : users){
		dest = user.get_id();
		rc = (sekey_error)sekey_find_key_v1(chosen, source, dest, se_key_type::symmetric_data_encryption);
		if(rc == SEKEY_OK){
			cout << "Key for " + source + ", " + dest + ": " + chosen << endl;
		} else {
			cout << "Key for " + source + ", " + dest + " not found. Returned value: " << rc << endl;
		}
	}

	/* now we simulate, using hard-coded values, the search for the most secure key to be used in specific
	 * combinations like (U1, U2), (U1, U2, U5), etc. */
	source = currentuser.userid;
	vector<string> dest_;
	if(currentuser.userid.compare("U1") == 0){
		dest_.clear();
		dest_.push_back("U4");
		dest_.push_back("U5");
		rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
		if(rc == SEKEY_OK){
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + ": " + chosen << endl;
		} else {
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + " not found. Returned value: " << rc << endl;
		}
		dest_.clear();
		dest_.push_back("U2");
		dest_.push_back("U5");
		rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
		if(rc == SEKEY_OK){
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + ": " + chosen << endl;
		} else {
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + " not found. Returned value: " << rc << endl;
		}
		dest_.clear();
		dest_.push_back("U4");
		dest_.push_back("U3");
		rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
		if(rc == SEKEY_OK){
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + ": " + chosen << endl;
		} else {
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + " not found. Returned value: " << rc << endl;
		}
	}
	if(currentuser.userid.compare("U2") == 0){
		dest_.clear();
		dest_.push_back("U1");
		dest_.push_back("U5");
		rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
		if(rc == SEKEY_OK){
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + ": " + chosen << endl;
		} else {
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + " not found. Returned value: " << rc << endl;
		}
	}
	if(currentuser.userid.compare("U3") == 0){
		dest_.clear();
		dest_.push_back("U1");
		dest_.push_back("U5");
		rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
		if(rc == SEKEY_OK){
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + ": " + chosen << endl;
		} else {
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + " not found. Returned value: " << rc << endl;
		}
		dest_.clear();
		dest_.push_back("U4");
		dest_.push_back("U5");
		rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
		if(rc == SEKEY_OK){
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + ": " + chosen << endl;
		} else {
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + " not found. Returned value: " << rc << endl;
		}
		dest_.clear();
		dest_.push_back("U1");
		dest_.push_back("U4");
		rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
		if(rc == SEKEY_OK){
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + ": " + chosen << endl;
		} else {
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + " not found. Returned value: " << rc << endl;
		}
	}
	if(currentuser.userid.compare("U4") == 0){
		dest_.clear();
		dest_.push_back("U1");
		dest_.push_back("U5");
		rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
		if(rc == SEKEY_OK){
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + ": " + chosen << endl;
		} else {
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + " not found. Returned value: " << rc << endl;
		}
		dest_.clear();
		dest_.push_back("U3");
		dest_.push_back("U5");
		rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
		if(rc == SEKEY_OK){
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + ": " + chosen << endl;
		} else {
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + " not found. Returned value: " << rc << endl;
		}
		dest_.clear();
		dest_.push_back("U1");
		dest_.push_back("U3");
		rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
		if(rc == SEKEY_OK){
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + ": " + chosen << endl;
		} else {
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + " not found. Returned value: " << rc << endl;
		}
		dest_.clear();
		dest_.push_back("U2");
		dest_.push_back("U3");
		rc = (sekey_error)sekey_find_key_v3(chosen, source, dest_, se_key_type::symmetric_data_encryption);
		if(rc == SEKEY_OK){
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + ": " + chosen << endl;
		} else {
			cout << "Key for " + source + " " + dest_[0] + " " + dest_[1] + " not found. Returned value: " << rc << endl;
		}
	}

	// stop SEkey and logout
	sekey_stop();
	cout << "SEkey stopped." << endl;
	l1->L1Logout();
	return;
}
